;; Generated package description from beframe.el  -*- no-byte-compile: t -*-
(define-package "beframe" "0.1.8" "Isolate buffers per frame" '((emacs "28.1")) :commit "6f9c6322a285bc75191853ac0c7e2fefff0ac354" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou General Issues" . "~protesilaos/general-issues@lists.sr.ht") :url "https://git.sr.ht/~protesilaos/beframe")
